#!/usr/bin/python

# Copyright (c) 1999-2002 Red Hat, Inc.  Distributed under GPL.
#
# Author: Adrian Likins <alikins@redhat.com>
#

import sys
sys.path.append("/usr/share/rhn/")
from up2date_client import up2dateErrors
from up2date_client import up2dateAuth
from up2date_client import up2dateUtils
from up2date_client import up2date
from up2date_client import config
from up2date_client import rpcServer
from up2date_client import rpmUtils
from up2date_client import transaction
import rpm
import os

cfg = config.initUp2dateConfig()

# mark this module as acceptable
__rhnexport__ = [
    'update',
    'remove',
    'refresh_list',
    'runTransaction',
    'verify'
]

argVerbose = 0  # verbose output (might clutter syslog)

global fd
def remove(pkgList):
    """We have been told that we should remove packages"""
    def rpmCallback(what, amount, total, hdr, path):
        if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
            global fd
            fileName = "%s/%s-%s-%s.%s.rpm" % (path,
                                               hdr['name'],
                                               hdr['version'],
                                               hdr['release'],
                                               hdr['arch'])
            try:
                 fd = os.open(fileName, os.O_RDONLY)
            except:
                print "Error opening %s for reading" % fileName
                return -1

            return fd
        elif what == rpm.RPMCALLBACK_INST_START:
            fileName = "%s-%s-%s.%s.rpm" % (hdr['name'],
                                            hdr['version'],
                                            hdr['release'],
                                            hdr['arch'])
        elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
            os.close(fd)

        elif what == rpm.RPMCALLBACK_INST_PROGRESS:
            pass
        elif what == rpm.RPMCALLBACK_UNINST_STOP:
            pass

    if type(pkgList) != type([]):
        return (13, "Invalid arguments passed to function",{})

    if argVerbose > 1:
        print "Called remove_packages(%s)" % (pkgList,)

    # FIXME: figure out the error handling
    try:
        up2date.removePackages(pkgList, rpmCallback)
    except up2dateErrors.DependencyError, e:
        # just return the list of deps if we failed
        return (14, "%s" % e, {})
    except up2dateErrors.RpmRemoveError, e:
        # if we later need finer grained info, we can
        return (15, "%s" % e, {})
    except up2dateErrors.RpmRemoveSkipListError, e:
        return (16, "%s could not be removed because it was on the skip list" % e, {})

    # if were succesful, return "4"
    return (0, "%s removed successfully" % pkgList, {})

global fd
def rpmCallback(what, amount, total, hdr, path):
    if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
        global fd
        fileName = "%s/%s-%s-%s.%s.rpm" % (path,
                                           hdr['name'],
                                           hdr['version'],
                                           hdr['release'],
                                           hdr['arch'])
        try:
             fd = os.open(fileName, os.O_RDONLY)
        except:
            print "Error opening %s for reading" % fileName
            return -1

        return fd
    elif what == rpm.RPMCALLBACK_INST_START:
        if type(hdr) == type(""):
            print "Repackaging..."
        else:
            fileName = "%s-%s-%s.%s.rpm" % (hdr['name'],
                                            hdr['version'],
                                            hdr['release'],
                                            hdr['arch'])
    elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
        os.close(fd)

    elif what == rpm.RPMCALLBACK_INST_PROGRESS:
        pass
    elif what == rpm.RPMCALLBACK_UNINST_STOP:
        pass


def noneToEmpty(value):
    if value == None:
        return ""
    return value

# stringing nay None's in dep tuples
def __cleanDeps(deps):
    clean_deps = []
    for dep in deps:
	( (name, version, release), (needsName, needsVersion),
          flags, suggested, sense) = dep

        values = [name, version, release, needsName, needsVersion, flags, suggested, sense]
        name, version,release, needsName, needsVersion, \
              flags, suggested, sense = map(noneToEmpty, values)
        
	clean_deps.append(((name,version,release),
                           (needsName, needsVersion),
                           flags, suggested, sense))
    return tuple(clean_deps)

def __cleanFailures(failures):
    clean_failures  = []
    for failure in failures:
        (msg, (flags, suggest, sense)) = failure
        values = [flags, suggest, sense]
        flags, suggest, sense = map(noneToEmpty, values)
        clean_failures.append((msg, (flags, suggest, sense)))
    return clean_failures

def __cleanConflicts(conflicts):

    clean_conflicts = []
    for conflict_index  in range(len(conflicts)):
        reason = conflicts[conflict_index][0]
        pkg = conflicts[conflict_index][1]
        tmp = list(pkg[:])
        for i in range(len(pkg)):
            if pkg[i] == None:
                tmp[i] = ""
        clean_conflicts.append((reason, tmp))
    return clean_conflicts

def update(pkgList):        
    """We have been told that we should retrieve/install packages"""

    def rpmCallback(what, amount, total, hdr, path):
        if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
            global fd
            fileName = "%s/%s-%s-%s.%s.rpm" % (path,
                                               hdr['name'],
                                               hdr['version'],
                                               hdr['release'],
                                               hdr['arch'])
            try:
                 fd = os.open(fileName, os.O_RDONLY)
            except:
                print "Error opening %s for reading" % fileName
                return -1

            return fd
        elif what == rpm.RPMCALLBACK_INST_START:
            fileName = "%s-%s-%s.%s.rpm" % (hdr['name'],
                                            hdr['version'],
                                            hdr['release'],
                                            hdr['arch'])
        elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
            os.close(fd)

        elif what == rpm.RPMCALLBACK_INST_PROGRESS:
            pass
        elif what == rpm.RPMCALLBACK_UNINST_STOP:
            pass

    if type(pkgList) != type([]):
        return (13, "Invalid arguments passed to function", {})

    if argVerbose > 1:
        print "Called update_packages(%s)" % (pkgList,)

    # if e,v,r is "", pretend it is version 0, release 0, epoch ""
    for pkg in pkgList:
        if pkg[1] == pkg[2] == pkg[3] == "":
            pkg[1] = pkg[2] = 0
            pkg[3] = ""

    try:
        ret = batchRun(0, pkgList)
    # catch all the exceptions we raise, and hopefully return a useful error
    # message to the server
    except up2dateErrors.RpmError, e:
        return (18, "Failed: packages requested raised dependency problems", {})
    except up2dateErrors.RpmInstallError, e:
        data = {}
        data['version'] = "0"
        data['name'] = "package.update.package_install_failure"
        if e.pkg:
            data = {'package_that_failed':pkg}
        return (32, "Failed: Packages failed to install properly: %s" % e.errmsg, data)
    except up2dateErrors.SkipListError, e:
	data = {}
	if e.pkglist:
		data = {'packages_on_skip_list': e.pkglist}
	# The version field is the version of the data struct
	data['version'] = "0"
	data['name'] = "package.update.packages_on_skip_list"
	return (30, "Failed: Some of the packages specified were on a skip list", data)
    except up2dateErrors.DependencyError, e:
	data = {}
	if e.deps:
	    # deps might have None's, cleanse them
	    data = {'failed_deps':__cleanDeps(e.deps)}
	data["version"] = "0"
	data["name"] = "package.update.failed_deps"
	return (18, "Failed: packages requested raised dependency problems", data)
    except up2dateErrors.UnsolvedDependencyError, e:
	data = {}
	if e.dep:
	   data = {'failed_deps':e.dep}
	if e.pkgs:
	   data["packages"] = e.pkgs
	   packages = up2dateUtils.pkglistToString(e.pkgs)
	else:
	   packages = ""
	data["version"] = "0"
	data["name"] = "package.update.deps_on_skip_list"
        return(29, "The depenecy %s is provided by the folowing packages: %s, but they are not available due to client configuration (skip lists, etc)" % (e.dep, packages), data)
    except up2dateErrors.ConflictError, e:
	if e.rc:
            # we cant marshall long ints, so make sure we cast them to ints
            try:
                tmprc = e.rc
                # old rpm always makes this an int, new rpm makes it a long int, we nee 
                if type(e.rc[0][1][2]) == type(1L): 
                    tmprc = [ (e.rc[0][0], (e.rc[0][1][0], e.rc[0][1][1], int(e.rc[0][1][2])) )]

            except TypeError:
                tmprc = "There was a parsing error with the rpm conflict message"
            except OverflowError:
                # must of be a real long int, this shouldnt ever happen in this case, but
                # dont traceback just in case
                tmprc = "There was an overflow error converting the return code to ints"
	    data = {'package_conflicts': tmprc}
        else:
	    data = {}
	data["version"] = "0"
	data["name"] = "package.update.package_conflicts"
	return (26, "Failed: There was a package conflict error with the package set: %s" % e, data)
    except up2dateErrors.FileConflictError, e:
	if e.rc: 
	    data = {'package_file_conflicts': e.rc}
	else:
	    data = {}
	data["version"] = "0"
	data["name"] = "package.update.package_file_conflicts"
        return (27, "Failed: There was file conflict error with the package set: %s" % e, data)
    except up2dateErrors.DependencySenseError, e:
	data = {}
	if e.sense:
		data = {'dependency_sense':e.sense}
	data["version"] = "0"
	data["name"] = "package.update.dependency_sense"
	return (28, "There was a dependency sense error with the sense: %s" % e.sense, data)
    except up2dateErrors.CommunicationError, e:
        return (21, "Failed: There was a communication error talking to the server", {})
    except up2dateErrors.FileNotFoundError, e:
        data = {'missing_package_error': e}
        data["version"] = "0"
        data["name"] = "package.update.file_not_found"
        return (31, "Failed: There was a File Not Found error. The error was: %s" % e, data)
    except up2dateErrors.GPGVerificationUnsignedPackageError, e:
	data = {'unsigned_package':e.pkg}
	data["version"] = "0"
	data["name"] = "package.update.unsigned_package"
        return (22, "Failed: There was a package gpg verification error. The error was: %s" % e, data)
    except up2dateErrors.GPGVerificationError, e:
	data = {'gpg_failed_package':e.pkg}
	data["version"] = 0
	data["name"] = "package.update.gpg_failed_package"
        return (23, "Failed: There was a package gpg verification error. The error was: %s" % e, {})
    except up2dateErrors.GPGInstallationError:
        return (24, "Failed: gpg is not properly installed on this machine", {})
    except up2dateErrors.GPGKeyringError:
        return (25, "Failed: The package signing key for Red Hat, Inc. is not on the gpg keyring", {})
    except up2dateErrors.OutOfSpaceError, e:
	data = {'space_required':e.totalSize}
	data['space_available'] = e.freeDiskSpace
	data['version'] = 0
	data['name'] = "package.up2date.out_of_space"
	return (33, "Failed: The was an error installing the package: the error was: %s" % e, data)
 
    # bit of a hack, but slightly "better messaging"
    if cfg.readEntry("retrieveOnly"):
	return (0, "Packages sucessfully retrieved but due to client specified options, the packages were downloaded, but not installed.", {})
    else:
    	return (0, "Packages were installed successfully", {})


def batchRun(onlyList, pkgList):
    from up2date_client import wrapper
    ret = wrapper.batchRun(onlyList, [], 0, fromDaemon = 1, actionPkgs = pkgList)
    return ret


# stringing nay None's in dep tuples
def __cleanDeps(deps):
    clean_deps = []
    for dep in deps:
	( (name, version, release), (needsName, needsVersion),
          flags, suggested, sense) = dep

        values = [name, version, release, needsName, needsVersion, flags, suggested, sense]
        name, version,release, needsName, needsVersion, \
              flags, suggested, sense = map(noneToEmpty, values)
        
	clean_deps.append(((name,version,release),
                           (needsName, needsVersion),
                           flags, suggested, sense))
    return tuple(clean_deps)

def __cleanFailures(failures):
    clean_failures  = []
    for failure in failures:
        (msg, (flags, suggest, sense)) = failure
        values = [flags, suggest, sense]
        flags, suggest, sense = map(noneToEmpty, values)
        clean_failures.append((msg, (flags, suggest, sense)))
    return clean_failures

def __cleanConflicts(conflicts):

    clean_conflicts = []
    for conflict_index  in range(len(conflicts)):
        reason = conflicts[conflict_index][0]
        pkg = conflicts[conflict_index][1]
        tmp = list(pkg[:])
        for i in range(len(pkg)):
            if pkg[i] == None:
                tmp[i] = ""
        clean_conflicts.append((reason, tmp))
    return clean_conflicts

# try to share this big old pile of exception handling:
def _catchErrors(method, name, args):
    ret = None
    
    try:
        ret = apply(method, args)
#    except up2dateErrors.RpmError, e:
#        return (18, "Failed: packages requested raised "\
#                "dependency problems", {})
    except up2dateErrors.RpmInstallError, e:
        data = {}
        data['version'] = "0"
        data['name'] = "%s.package_install_failure" % name
        if e.pkg:
            data = {'package_that_failed':e.pkg}
        return (32, "Failed: Packages failed to install "\
                "properly: %s" % e.errmsg, data)
    except up2dateErrors.SkipListError, e:
	data = {}
	if e.pkglist:
		data = {'packages_on_skip_list': e.pkglist}
	# The version field is the version of the data struct
	data['version'] = "0"
	data['name'] = "%s.packages_on_skip_list" % name
	return (30, "Failed: Some of the packages specified were "\
                "on a skip list", data)
    except up2dateErrors.DependencyError, e:
	data = {}
	if e.deps:
	    # deps might have None's, cleanse them
	    data = {'failed_deps':__cleanDeps(e.deps)}
	data["version"] = "0"
	data["name"] = "%s.failed_deps" % name
	return (18, "Failed: packages requested raised "\
                "dependency problems", data)
    except up2dateErrors.TransactionError, e:
	data = {}
	if e.deps:
	    # deps might have None's, cleanse them
            data = {'failed_elements': __cleanFailures(e.deps)}
#	    data = {'failed_deps':__cleanDeps(e.deps)}
	data["version"] = "0"
	data["name"] = "%s.failed_transaction" % name
        # FIXME: pretty up the test string error
	return (40, "Failed: The transaction failed for the following reasons:\n%s" %
                data['failed_elements'],
                data)
    except up2dateErrors.UnsolvedDependencyError, e:
	data = {}
	if e.dep:
	   data = {'failed_deps':e.dep}
	if e.pkgs:
	   data["packages"] = e.pkgs
	   packages = up2dateUtils.pkglistToString(e.pkgs)
	else:
	   packages = ""
	data["version"] = "0"
	data["name"] = "%s.deps_on_skip_list" % name
        return(29, "The dependency %s is provided by the folowing "\
               "packages: %s, but they are not available due to client "\
               "configuration (skip lists, etc)" % (e.dep, packages), data)
    except up2dateErrors.ConflictError, e:
	if e.rc:
            # we cant marshall long ints, so make sure we cast them to ints
            try:
                tmprc = e.rc
                # old rpm always makes this an int, new rpm makes it a
                # long int, we nee
                if type(e.rc[0][1][2]) == type(1L): 
                    tmprc = [ (e.rc[0][0], (e.rc[0][1][0], e.rc[0][1][1],
                                            int(e.rc[0][1][2])) )]

            except TypeError:
                tmprc = "There was a parsing error with the rpm "\
                        "conflict message"
            except OverflowError:
                # must of be a real long int, this shouldnt ever
                # happen in this case, but dont traceback just in case
                tmprc = "There was an overflow error converting "\
                        "the return code to ints"
	    data = {'package_conflicts': __cleanConflicts(tmprc)}
        else:
	    data = {}
	data["version"] = "0"
	data["name"] = "%s.package_conflicts" % name
	return (26, "Failed: There was a package conflict error "\
                "with the package set: %s" % e, data)
    except up2dateErrors.FileConflictError, e:
	if e.rc: 
	    data = {'package_file_conflicts': e.rc}
	else:
	    data = {}
	data["version"] = "0"
	data["name"] = "%s.package_file_conflicts" % name
        return (27, "Failed: There was file conflict error with the "\
                "package set: %s" % e, data)
    except up2dateErrors.DependencySenseError, e:
	data = {}
	if e.sense:
		data = {'dependency_sense':e.sense}
	data["version"] = "0"
	data["name"] = "%s.dependency_sense" % name
	return (28, "There was a dependency sense error with "\
                "the sense: %s" % e.sense, data)
    except up2dateErrors.CommunicationError, e:
        return (21, "Failed: There was a communication error "\
                "talking to the server", {})
    except up2dateErrors.FileNotFoundError, e:
        data = {'missing_package_error': e}
        data["version"] = "0"
        data["name"] = "%s.file_not_found" % name
        return (31, "Failed: There was a File Not Found error. "\
                "The error was: %s" % e, data)
    except up2dateErrors.GPGVerificationUnsignedPackageError, e:
	data = {'unsigned_package':e.pkg}
	data["version"] = "0"
	data["name"] = "%s.unsigned_package" % name
        return (22, "Failed: There was a package gpg verification error. "\
                "The error was: %s" % e, data)
    except up2dateErrors.GPGVerificationError, e:
	data = {'gpg_failed_package':e.pkg}
	data["version"] = 0
	data["name"] = "%s.gpg_failed_package" % name
        return (23, "Failed: There was a package gpg verification error. "\
                "The error was: %s" % e, {})
    except up2dateErrors.GPGInstallationError:
        return (24, "Failed: gpg is not properly installed "\
                "on this machine", {})
    except up2dateErrors.GPGKeyringError:
        return (25, "Failed: The package signing key for Red Hat, Inc. "\
                "is not on the gpg keyring", {})
    except up2dateErrors.PackageNotAvailableError, e:
        data = {}
        data['missing_packages'] = e.missing_packages
        data['version'] = 0
        data['name'] = "%s.packages_not_available" % name
        return (41, "Failed: Some of the packages requested are not available.", data)
    except up2dateErrors.OutOfSpaceError, e:
	data = {'space_required':e.totalSize}
	data['space_available'] = e.freeDiskSpace
	data['version'] = 0
	data['name'] = "%s.out_of_space" % name
	return (33, "Failed: The was an error installing the package: "\
                "the error was: %s" % e, data)
    return ret

def runTransaction(transactionData):
                                                                                                                                                           
    # build a transaction.TransactionData object
    tsd = transaction.TransactionData()
    tsd.data = transactionData

    # pass it to genTransaction to get a real rpm transaction
    try:
        ts = up2date.genTransaction(tsd)
    except up2dateErrors.PackageNotAvailableError, e:
        data = {}
        data['missing_packages'] = e.missing_packages
        data['version'] = 0
        data['name'] = "packages.runTransaction..packages_not_available"
        return (41, "Failed: Some of the packages requested are not available", data)

#    ret = up2date.runTransaction(ts, rpmCallback)
    ret = _catchErrors(up2date.runTransaction, "packages.runT ransaction", ( ts, rpmCallback))

    # little odd, if it returns a tuple, it's an error
    if type(ret) == type(()):
        return ret                                                                                                                                                           
    return (0, "packages.transaction ran successfully", {})


# push again the list of rpm packages to the server
def refresh_list():
    s = rpcServer.getServer()

    if argVerbose > 1:
        print "Called refresh_rpmlist"

    ret = None
    try:
        s.registration.update_packages(up2dateAuth.getSystemId(),
                                       rpmUtils.getInstalledPackageList())
    except:
        print "ERROR: refreshing remote package list for System Profile"
        return (20, "Error refreshing package list", {})

    return (0, "rpmlist refreshed", {})

def verify(packages):
    if argVerbose > 1:
        print "Called verify"

    data = {}
    data['name'] = "packages.verify"
    data['version'] = 0
    ret = rpmUtils.verifyPackages(packages)

    data['verify_info'] = ret
    return (0, "packages verified", data)

def verifyAll():
    if argVerbose > 1:
        print "Called verifyAll"
                                                                                
    data = {}     
    data['name'] = "packages.verifyAll"
    data['version'] = 0
    ret = rpmUtils.verifyAllPackages()
    data['verify_info'] = ret
    return (0, "packages verified", data)


def main():
#    print refresh_list()

    print verify(["pam", "glibc", "up2date"])
    sys.exit()

    up2dateAuth.updateLoginInfo()
    #packages1 = [(('wget', '1.8.2', '5', ''), 'u'),
    #             (('cvs', '1.11.2', '8', ''), 'u')]
                                                                                                                                                           
    #packages2 = [(('glibc', '2.3.2', '86', '', "i686"), 'u')]
    packages3 = [(('zip', '2.3', '10', '',''), 'i'),
		  (('kernel', '2.4.9', 'e.10', '', ''), 'e')]
#		  (('pstack', '1.1' ,'1', '',''), 'e')]

    packages4 = [(('xinetd', '2.3.11', '2.AS2.1', '',''), 'e'),
		 (('xinetd', '2.3.3', '1', '',''), 'i')]
	
    packages5 = [(('xinetd', '2.3.11', '2.AS2.1', '',''), 'i'),
                 (('xinetd', '2.3.3', '1', '',''), 'e')]
	  
    data = {'packages':packages4}
                                                                                                                                                           
    print runTransaction(data)

if __name__ == "__main__":
    main()
