#!/usr/bin/python -u
#
# Client code for Update Agent
# Copyright (c) 1999-2002 Red Hat, Inc.  Distributed under GPL.
#
# Author: Preston Brown <pbrown@redhat.com>
#         Adrian Likins <alikins@redhat.com
#         Cristian Gafton <gafton@redhat.com>
#
"""Guts of the up2date client. This module provides abstration
of all the remote rpc calls that the client needs to make, includes
the dependency solve, and other assorted bits."""

import os,sys
import string
import fnmatch, glob
import rpm
#import rhnxmlrpclib as xmlrpclib
import up2dateUtils
import up2dateLog
import up2dateErrors
import up2dateAuth
import rpmSource
import rhnPackageInfo
import rhnChannel
import checkbootloader
import rpmUtils
import rpcServer
import depSolver
import headers
#import elilocfg
#import lilocfg
#import grubcfg

# so we can use the same code with python 1.5.2 and 2.2
from rhn import rpclib

# ugh, ugly kluge hack...
# we want to import the "new" versions, but 
# not much point in changing all the code since
# it's only for one specified set of versions
#import newelilocfg
elilocfg = __import__("up2date_client.newelilocfg", globals(), locals(), ['newelilocfg'])
lilocfg = __import__('up2date_client.newlilocfg', globals(), locals(), ['newlilocfg'])
grubcfg = __import__("up2date_client.newgrubcfg", globals(), locals(), ['newgrubcfg'])

from translate import _, N_, cat

import config

cfg = config.initUp2dateConfig()

def reloadConfig():
    cfg.load()

def version():
    # substituted to the real version by the Makefile at installation time.
    return "@VERSION@"


DEBUG=0
def DPRINT(string):
    if DEBUG:
        print  string

# get around cached dependencies.  Gross ugly hack.
rpm.addMacro("_dbi_tags","Packages:Name:Basenames:Group:Requirename:Providename:Conflictname:Triggername")

# this is a "gloabal" variable caching the list of all available packages
global avail_package_list
avail_package_list = []

try:
    log = up2dateLog.initLog()
except IOError, e:
    print "Unable to open log file. The error was: %s" % e
    print _("You must run the Update Agent as root.")
    sys.exit(1)



def checkForStorageDir():
    dir = cfg.readEntry("storageDir")
    if os.access(dir, os.X_OK):
        return 0
    else:
        return 1

    

headerCache = headers.HeaderCache()



def getAvailablePackageList(msgCallback=None, progressCallback=None):
    global avail_package_list


    if len(avail_package_list):
        return avail_package_list

    package_list = rhnPackageInfo.availablePackageList(msgCallback, progressCallback)

    availPkgsDict = {}
    #tmp_args,tmp_method = xmlrpclib.loads(data)
    for pkg in package_list:
        archScore = rpm.archscore(pkg[4])
        if archScore <= 0:
            # incompatible architecture
            continue

        pkgName = pkg[0]
        if not availPkgsDict.has_key(pkgName):
            availPkgsDict[pkgName] = pkg
            continue

        # Already seen this name
        ret = up2dateUtils.comparePackages(availPkgsDict[pkgName], pkg)
        if ret > 0:
            # don't care, we already have a better version
            continue
        if ret < 0:
            # Better version
            availPkgsDict[pkgName] = pkg
            continue
        # Same version, so compare the arches
        oldScore = rpm.archscore(availPkgsDict[pkgName][4])
        if oldScore <= archScore:
            continue
        # Package with a better architecture
        availPkgsDict[pkgName] = pkg

    # Sort the keys (i.e. the package names)
    avail_package_list = availPkgsDict.values()
    avail_package_list.sort()
    return avail_package_list

all_avail_package_lst = []
def allAvailablePackageList(msgCallback=None, progressCallback=None):
    log = up2dateLog.initLog()
    global all_avail_package_lst
    # if this is non-zero, we've already fetched it
    if len(all_avail_package_lst):
        return all_avail_package_lst

    availPkgList = []

    try:
        log.log_me("allAvailablePackageList from network")
    except:
        pass


    # FIXME: replace with a up2date.getChannels() equilvinet when
    #        I find out where it belongs
    channels = rhnChannel.getChannels()
    if not channels:
        return None
    availPkgList = []

    new_channels = []
    for channel in channels:
	if channel[0] not in rhnChannel.channel_blacklist:
	    new_channels.append(channel)
    channels = new_channels
    
    for channel in channels:
        if msgCallback:
            msgCallback("Fetching all package list for channel: %s" % channel[0] )
        # a channel maybe empty, with no version
        if not channel[1]:
            continue
        package_list,type = rpcServer.doCall(
            rpmSource.psc.listPackageSource.listAllPackages,
            channel[0],
            channel[1],
            msgCallback = msgCallback,
            progressCallback = progressCallback )


        availPkgList = availPkgList + package_list

    all_avail_package_lst = availPkgList
    return all_avail_package_lst



all_avail_package_list = []
def getAllAvailablePackageList(msgCallback=None, progressCallback=None):
    global all_avail_package_list
    
    if len(all_avail_package_list):
        return all_avail_package_list

    package_list = allAvailablePackageList(
        msgCallback, progressCallback)

    pkgDict = {}
    for pkg in package_list:
        archScore = rpm.archscore(pkg[4])
        if archScore <= 0:
            # incompatible architecture
            continue

        nvre = tuple(pkg[0:3])
        # first package that matches, no need to test anything else
        if not pkgDict.has_key(nvre):
            pkgDict[nvre] = pkg
            continue

        oldScore = rpm.archscore(pkgDict[nvre][4])
        if oldScore <= archScore:
            continue

        # else this is a better arch, so use it...
        pkgDict[nvre] = pkg
                   
    # should have a dict with the best
    pkg_list = pkgDict.values()
    all_avail_package_lst = package_list
    return all_avail_package_lst

def getPackageSearchPath():
    dir_list = []
    dir_list.append(cfg.readEntry("storageDir"))

    dir_string = cfg.readEntry("packageDir")
    if dir_string: 
    	dir_list = dir_list + string.split(dir_string, ':')
    return dir_list


# needed for actions, maybe client to at some point
def getErrataInfo(errata_id):
    s = rpcServer.getServer()
    packages = rpcServer.doCall(s.errata.getErrataInfo, up2dateAuth.getSystemId(), errata_id)
    
    return packages


# pass the lite flag if you only want the "lite" header
def getHeader(pkg,lite=None):
    hdr = None

    hdr,type = rpcServer.doCall(rpmSource.psc.headerSource.getHeader, pkg)
    return hdr
    
    

# implicitly places headers in the headerCache.
# FIXME: i think this can die now with headers.HeaderList
def getHeaders(pkgList, msgCallback = None, progressCallback = None,
               refreshCallback = None):

    headerList = []
        
    # if we get a single header, listify it
    # double check to make sure it's a list of lists and not just
    # a single list item
    if type(pkgList) == type([]):
        if len(pkgList):
            if type(pkgList[0]) == type(""):
                tmp = [pkgList]
                pkgList = tmp
        else:
            return []
    else:
        raise TypeError, "pkgList must be a list of lists"


    # munge the list if we have the size include.
    # xml-rpc doesnt seem fond of that being included
    total = len(pkgList)

    # reset progress bar
    if progressCallback:
        progressCallback(0, total)

    if not refreshCallback:
        refreshCallback = 0

    for pkg in pkgList:
        if refreshCallback:
            refreshCallback()
        hdr = None
        hdr, sourcetype = rpcServer.doCall(rpmSource.psc.headerSource.getHeader, pkg)
        
        headerList.append(hdr)

        if progressCallback:
            progressCallback(len(headerList), total)
 
    return headerList

def removeHeader(pkg):
    # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.hdr" %
                          (cfg.readEntry("storageDir"),
                           up2dateUtils.pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)

    fileNames = glob.glob("%s/%s.noarch.hdr" %
                          (cfg.readEntry("storageDir"),
                           up2dateUtils.pkgToString(pkg)))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)


def isPackageCached(pkg):

    baseFileName = "%s-%s-%s.%s.rpm" % (pkg[0], pkg[1], pkg[2], pkg[4])

    fileName = "%s/%s" % (cfg.readEntry("storageDir"),
                          baseFileName)

    if (os.access(fileName, os.R_OK)):
        size = os.stat(fileName)[6]
        if size == long(pkg[5]):
            return 1
    return 0


def getPackage(pkg, msgCallback = None, progressCallback = None):

    buffer = rpcServer.doCall(rpmSource.psc.packageSource.getPackage, pkg, msgCallback, progressCallback)

    # retrieve source packages as well
    if cfg.readEntry('retrieveSource'):
        hdr = getHeader(pkg)
        srcrpm = hdr['sourcerpm']
        channel = pkg[6]
        buffer = rpcServer.doCall(rpmSource.psc.sourcePackageSource.getPackageSource,
                        channel,
                        srcrpm,
                        msgCallback,
                        progressCallback)
        
    return 0


def removePackage(pkg):
     # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.*" %
                          (cfg.readEntry("storageDir"),
                           up2dateUtils.pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
            
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)
        
    fileNames = glob.glob("%s/%s.noarch.*" %
                          (cfg.readEntry("storageDir"),
                           up2dateUtils.pkgToString(pkg)))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)


def avail_package_dict():
    # returns a hash version of the Avialable Package List
    availablePkgList = getAvailablePackageList()
    pkgDict = {}
    for pkg in availablePkgList:
        pkgDict[up2dateUtils.pkgToStringArch(pkg)] = pkg
    return pkgDict


def getArchesOfAvailablePackages(availablePkgList,pkgName):
    archList = []
    for pkg in availablePkgList:
        if pkgName == pkg[0]:
            archList.append(pkg[4])

    return archList

# convience class to make removeInstalledPackagesFromList cleaner
class UpdateList:
    def __init__(self):
        self.updateDict = {}
        self.__updateIndex = None

    def nameList(self):
        return self.updateDict.keys()

    # gen a list of just the name,version,release,epoch, and arch fields
    def nvreaList(self):
        return map(lambda a: a[:5], self.updatedList())

    def nvreList(self):
        return map(lambda a: a[:4], self.updatedList())

    def appendIfNewer(self, installedPkg, pkg):
        pkgName = pkg[0]
        # do we already have a package like this?
        if not self.updateDict.has_key(pkgName):
            self.updateDict[pkgName] = pkg
            return
        # do we have a match for name-version-release
        if pkg[:4] in self.nvreList():
            # see if this is a better matching architecture
	    archScore = rpm.archscore(pkg[4])
            # incompatible arch
            if archScore <= 0:
                return
	    oldScore = rpm.archscore(self.updateDict[pkgName][4])
            if archScore < oldScore: # retain the better arch
                self.updateDict[pkgName][4] = pkg[4]
            return
        # else, we might gave an older one that needs updating
        ret = up2dateUtils.comparePackages(self.updateDict[pkgName],pkg)
        if ret > 0: # if pkg is newer
            self.updateDict[pkgName] = pkg

    def updatedList(self):
        return self.updateDict.values()

    def delete(self, pkg):
        del self.updateDict[pkg[0]]



# check if the signature matches the redhat.com one
def hasBadSignature(pkg):
    hdr = getHeader(pkg)
    fileName = "%s/%s.%s.rpm" % (cfg.readEntry("storageDir"),
                                 up2dateUtils.pkgToString(pkg),
                                 hdr['arch'])

    # all the gory bits of checking the sig is in rpmUtils now
    return rpmUtils.checkRpmSig(fileName)



# The new dryRun
def dryRun(pkgList, msgCallback = None, progressCallback = None, refreshCallback=None):
    # XXX: initialize correctly from the --verbose cmd line flag.
    debug = 0
    
    if debug:
        print "Called dryRun:", pkgList
        
    availPkgs = getAvailablePackageList()
    depsolve = depSolver.DependencySolver(pkgList, availPkgs)

    # set the callbacks wwe need
    depsolve.set_callback("msg", msgCallback)
    depsolve.set_callback("progress", progressCallback)
    depsolve.set_callback("refresh", refreshCallback)
    
    # set up the trabsaction set & all
    depsolve.setup()
    while 1:
        if debug:
            print "\nSelected so far", depsolve.selected
        ret = depsolve.solvedep()
        if debug:
            print "Got code %d from solvedep" % ret
        if ret == 0:
            break
        elif ret < 0: # not really reached
            raise up2dateErrors.DependencyError(_("Could not complete dependency resolution"))
    rc = depsolve.run()
    if progressCallback:
        progressCallback(1, 1)
    if rc and len(rc):
        msg = ""
        for code in rc:
            msg = msg + code[0] + "\n"
            raise up2dateErrors.ConflictError(_("Test install failed because of package conflicts:\n") + msg, rc)
    newList = []
    for s in depsolve.selected:
        if not s in pkgList:
            pkgList.append(s)
            newList.append(s)
    return (newList, [])

def getAllPkgInfo(allPkgList, pkg):
    pkgArchs = []
    for package in allPkgList:
        if list(pkg)[:4] == package[:4]:
            if pkg[4] != "":
                if list(pkg[:5]) == package[:5]:
                    return package
            else:
                pkgArchs.append(package)
#    print pkgArchs
    if pkgArchs:
        pkgArchs.sort(up2dateUtils.comparePackages)
#        print pkgArchs
        pkgArchs.reverse()
        return pkgArchs[0]

    
    return None

def genTransaction(transactionData):
    # generate a rpm.TransactionSet from a TransactionData object
    #

    allPkgs = getAllAvailablePackageList()

    db = rpmUtils.openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)
    
    pkgList = transactionData.data['packages']

    kernels = []
    packagesToGet = []
    missing_packages = []
    packagesNotInstalled = []

    
    for (pkg,mode) in pkgList:
        # pkgList already picks out the best arch for the
        #system...
        
        # add it to the transaction setup approriately
        if mode == 'u' or mode == 'i':
            package = getAllPkgInfo(allPkgs, pkg)
            if not package:
                missing_packages.append(pkg)
                continue
        
            hdr = getHeader(package)
            packagesToGet.append(package)
            if not hdr:
                raise up2dateErrors.RpmError(_(
                    "Could not find header for %s") % pkg)
            if "kernel" in hdr['Providename']:
                kernels.append(hdr)
            ts.add(hdr,hdr, mode)
        elif mode == "e":
            # jbj thinks there might be issues with adding
            # a remove to the ts twice, so we may have to
            # be paranoid about that to avoid issues
            pkglabel = "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])
            try:
                ts.remove(pkglabel)
            except rpm.error, e:
                if e == "package not installed":
                    packagesNotInstalled.append(pkglabel)
            # FIXME
            # we skip a package to remove if its not installed,
            # ideally we should throw an error, but for not, it's
            # better to be consistent with deployed code
                
        else:
            ts.add(hdr, hdr, mode)

    if missing_packages:
        errmsg = _("Some of the packages in this transaction are not available.")
        raise up2dateErrors.PackageNotAvailableError(errmsg, missing_packages)


    # rpm 4.0 doesnt support this
##    if transactionData.data.has_key('vsflags'):
##        vsflag = 0
##        for flag in transactionData.data['vsflags']:
##            vsflag = vsflag | flag
##        ts.setVSFlags(vsflag)

    # should probabaly dep check before we actually
    # download the package

    for package in packagesToGet:
        # FIXME: progress callbacks...
       getPackage(package)

    return ts
    
   
def runTransaction(ts, rpmCallback, transdir=None, rollbacktrans=1):
    if transdir == None:
        transdir = cfg.readEntry('storageDir')
    # reset the flags, but save the current ones
    oldInstalledPackages = rpmUtils.getInstalledPackageList()

    ts.order()
    deps = ts.depcheck()
    
    if deps:
        raise up2dateErrors.DependencyError(_(
            "Dependencies should have already been resolved, "\
            "but they are not."), deps)


    
    if cfg.readEntry("justdb"):
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        # ( 1 << 10) == rpm.RPMTRANS_FLAG_REPACKAGE
	# hmm, this version of rpm gripes about deltas without
	# this
	transflag = 0
	if cfg.readEntry("enableRollbacks"):
	    transflag = (1 << 10)
        rc = ts.run(transflag, rpm.RPMPROB_FILTER_OLDPACKAGE , rpmCallback, cfg.readEntry("storageDir"))

    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise up2dateErrors.TransactionError(_(
            "Failed running transaction of  packages: %s") % errors, deps=rc)


    newInstalledPackages = rpmUtils.getInstalledPackageList()

    removedPackages = []
    addedPackages = []

    # find packages that were removed
    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    for pkg in newInstalledPackages:
        if pkg not in oldInstalledPackages:
            addedPackages.append(pkg)

    if len(removedPackages):
    	rhnPackageInfo.remoteDelPackages(removedPackages)
    if len(addedPackages):    
    	rhnPackageInfo.remoteAddPackages(addedPackages)

            
                 
def installPackages(pkgList, rpmCallback):
    try:
        log.log_me("installing packages: %s " % up2dateUtils.pprint_pkglist(pkgList))
    except:
        pass

    # a list of all the packge names to treat as kernels.
    # "hardcode" it here so people dont mess with it as much
    total = len(pkgList)

    oldInstalledPackages = rpmUtils.getInstalledPackageList()
    
    # delete any existing db, open in write mode
    db = rpmUtils.openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)

    kernels = []
    for p_i in range(total):
        hdr = getHeader(pkgList[p_i])
        if hdr:
            if "kernel" in hdr['provides']:
                ts.add(hdr,hdr, "i")
                kernels.append(hdr)
            else:
                ts.add(hdr, hdr, "u")
        else:
            raise up2dateErrors.RpmError(_("Could not find header for %s") % pkgList[p_i])

    ts.order()
    deps = ts.depcheck()
    if deps:
        raise up2dateErrors.DependencyError(_("Dependencies should have already been resolved, but they are not."), deps)

    if cfg.readEntry("justdb"):
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        # ( 1 << 10) == rpm.RPMTRANS_FLAG_REPACKAGE
	transflag = 0
	if cfg.readEntry("enableRollbacks"):
	    transflag = (1 << 10)
        rc = ts.run(transflag , 0, rpmCallback, cfg.readEntry("storageDir"))
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise up2dateErrors.RpmError(_("Failed installing packages: %s") % errors)

    newInstalledPackages = rpmUtils.getInstalledPackageList()

    removedPackages = []
    addedPackages = []

    # find packages that were removed
    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    for pkg in newInstalledPackages:
        if pkg not in oldInstalledPackages:
            addedPackages.append(pkg)

    if len(removedPackages):
    	rhnPackageInfo.remoteDelPackages(removedPackages)
    if len(addedPackages):    
    	rhnPackageInfo.remoteAddPackages(addedPackages)

    return kernels

def removePackages(pkgList, rpmCallback):
    try:
        log.log_me("removing packages: %s " % up2dateUtils.pprint_pkglist(pkgList))
    except:
        pass  

    total = len(pkgList)

    oldInstalledPackages = rpmUtils.getInstalledPackageList()

    # delete any existing db, open in write mode
    db = rpmUtils.openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)

    # FIXME: just that a package is installed before claiming
    # it cant be removed

    remove_list = []
    for p_i in range(total):
        for pattern in cfg.readEntry("removeSkipList"):
            pkgName = "%s-%s-%s" % (pkgList[p_i][0],pkgList[p_i][1],pkgList[p_i][2])
            # check the package name against the globs in removeSkipList
            if fnmatch.fnmatch(pkgList[p_i][0], pattern):
                raise up2dateErrors.RpmRemoveSkipListError, pkgName
            else:
                remove_list.append(pkgName)

    failed_packages = {}
    for pkgName in remove_list:
        try:
            ts.remove(pkgName)
        except rpm.error, e:
            failed_packages[pkgName] = e

    # if we get failed packages, raise the hash
    # of failed packages
    if len(failed_packages.keys()):
        raise up2dateErrors.RpmRemoveError, failed_packages
    
    ts.order()
    deps = ts.depcheck()
    
    # FIXME: return failed deps if any
    if deps:
        raise up2dateErrors.DependencyError, deps

    if cfg.readEntry("justdb"):
        # do we use storage dir here?? do we need it at all?
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        rc = ts.run(0, 0, rpmCallback, cfg.readEntry("storageDir"))
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise up2dateErrors.RpmError(_("Failed removing packages: %s") % errors)

    newInstalledPackages = rpmUtils.getInstalledPackageList()
    removedPackages = []
    # find packages that were removed

    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    if len(removedPackages):
    	rhnPackageInfo.remoteDelPackages(removedPackages)
        log.log_me("Removed packages: %s" % removedPackages)

    
    return 0

def installBootLoader(kernels):
    
    # option to disable this permanently for those using grub,etc
    if cfg.readEntry("noBootLoader"):
        return

    # dont run the new lilo if were running "justdb"
    if cfg.readEntry("justdb"):
        return

    log.log_me("Modifying bootloader config to include the new kernel info")
    kernelList = []
    for krn in kernels:
        extraInfo = krn['name']
        
        # this logics a bit weird
        if extraInfo == None:
            infoString = ""
        elif extraInfo == "kernel":
            infoString = ""
        else:
            # if its a kernel is named like "kernelsdfasd" the
            # info blurb is gonna be "kernelsdfasd". I can live
            # with that. 
            infoString = string.split(extraInfo, "-", 1)[-1:][0]
            # just in case some kook names a kernel "kernel-"
            if infoString == "":
                infoString = extraInfo
                


        verRel = "%s-%s%s" % (krn['version'], krn['release'],infoString)
        kernelList.append((verRel,extraInfo))
        log.log_me("Adding %s to bootloader config" % verRel)


    bootloader = checkbootloader.whichBootLoader()
    
    if up2dateUtils.archGlob() == 'i?86' and bootloader == "LILO":
        __install_lilo(kernelList)
    elif up2dateUtils.archGlob() == 'i?86' and bootloader == "GRUB":
        # at the moment, kernel rpms are supposed to grok grub
        # and do the Right Thing. Just a placeholder for doc purposes
        __install_grub(kernelList)
    elif up2dateUtils.archGlob() == "ia64":
        __install_elilo(kernelList)

def __install_elilo(kernelList):
    log.log_me("Running elilo with the new configuration")
    ret = elilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_lilo(kernelList):
    log.log_me("Running lilo with the new configuration")
    ret = lilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_grub(kernelList):
    log.log_me("Installing the kernel via grub")
    ret  = grubcfg.installNewImages(kernelList,test=0)
    return ret

